<?php

namespace BuilderiusPro\Bundle\BuilderBundle\Helper;

class UserSwitchingHelper
{
    /**
     * @param \WP_User $user A WP_User object (usually from the logged_in cookie).
     * @return bool Whether verification with the auth cookie passed.
     */
    public static function authenticateOldUser(\WP_User $user)
    {
        $cookie = self::getAuthCookie();
        if (!empty($cookie)) {
            if (self::secureAuthCookie()) {
                $scheme = 'secure_auth';
            } else {
                $scheme = 'auth';
            }

            $old_user_id = wp_validate_auth_cookie(end($cookie), $scheme);

            if ($old_user_id) {
                return ($user->ID === $old_user_id);
            }
        }
        return false;
    }

    /**
     * @param int $user_id The ID of the user to switch to.
     * @param bool $remember Optional. Whether to 'remember' the user in the form of a persistent browser cookie. Default false.
     * @param bool $set_old_user Optional. Whether to set the old user cookie. Default true.
     * @param bool $clear_old_user Optional. Whether to clear the old user cookie. Default true.
     * @return false|\WP_User WP_User object on success, false on failure.
     */
    public static function switchToUser($user_id, $remember = false, $set_old_user = true, $clear_old_user = false)
    {
        $user = get_userdata($user_id);

        if (!$user) {
            return false;
        }

        $old_user_id = (is_user_logged_in()) ? get_current_user_id() : false;
        $old_token = function_exists('wp_get_session_token') ? wp_get_session_token() : '';
        $auth_cookies = self::getAuthCookie();
        $auth_cookie = end($auth_cookies);
        $cookie_parts = $auth_cookie ? wp_parse_auth_cookie($auth_cookie) : false;

        if ($set_old_user && $old_user_id) {
            // Switching to another user
            $new_token = '';
            self::setOldUserCookie($old_user_id, false, $old_token);
        } elseif ($clear_old_user) {
            // Switching back, either after being switched off or after being switched to another user
            $new_token = ($cookie_parts && isset($cookie_parts['token'])) ? $cookie_parts['token'] : '';
            self::clearOldUserCookie(false);
        }

        /**
         * Attaches the original user ID and session token to the new session when a user switches to another user.
         *
         * @param array $session Array of extra data.
         * @param int $user_id User ID.
         * @return array Array of extra data.
         */
        $session_filter = function (array $session, $user_id) use ($old_user_id, $old_token) {
            $session['switched_from_id'] = $old_user_id;
            $session['switched_from_session'] = $old_token;
            return $session;
        };

        add_filter('attach_session_information', $session_filter, 99, 2);

        wp_clear_auth_cookie();
        wp_set_auth_cookie($user_id, $remember, '', $new_token);
        wp_set_current_user($user_id);

        remove_filter('attach_session_information', $session_filter, 99);

        if ($set_old_user) {
            do_action('builderius_switch_to_user', $user_id, $old_user_id, $new_token, $old_token);
        } else {
            do_action('builderius_switch_back_user', $user_id, $old_user_id, $new_token, $old_token);
        }

        if ($old_token && $old_user_id && !$set_old_user) {
            // When switching back, destroy the session for the old user
            $manager = \WP_Session_Tokens::get_instance($old_user_id);
            $manager->destroy($old_token);
        }

        return $user;
    }

    /**
     * @return bool
     */
    public static function switchOffUser()
    {
        $current_user_id = get_current_user_id();
        if (!$current_user_id) {
            return false;
        }

        if (!UserSwitchingHelper::getOldUser()) {
            $current_token = function_exists('wp_get_session_token') ? wp_get_session_token() : '';
            self::setOldUserCookie($current_user_id, false, $current_token);
        }
        wp_clear_auth_cookie();
        wp_set_current_user(0);

        /**
         * @param int $current_user_id The ID of the user switching off.
         * @param string $current_token The token of the session of the user switching off.
         */
        do_action('builderius_switch_off_user', $current_user_id, $current_token);

        return true;
    }

    /**
     * @param int $old_user_id The ID of the originating user, usually the current logged in user.
     * @param bool $pop Optional. Pop the latest user off the auth cookie, instead of appending the new one. Default false.
     * @param string $token Optional. The old user's session token to store for later reuse. Default empty string.
     * @return void
     */
    public static function setOldUserCookie($old_user_id, $pop = false, $token = '')
    {
        $secure_auth_cookie = self::secureAuthCookie();
        $secure_olduser_cookie = self::secureOldUserCookie();
        $expiration = time() + 172800; // 48 hours
        $auth_cookie = self::getAuthCookie();
        $olduser_cookie = wp_generate_auth_cookie($old_user_id, $expiration, 'logged_in', $token);

        if ($secure_auth_cookie) {
            $auth_cookie_name = BUILDERIUS_USER_SWITCHING_SECURE_COOKIE;
            $scheme = 'secure_auth';
        } else {
            $auth_cookie_name = BUILDERIUS_USER_SWITCHING_COOKIE;
            $scheme = 'auth';
        }

        if ($pop) {
            array_pop($auth_cookie);
        } else {
            array_push($auth_cookie, wp_generate_auth_cookie($old_user_id, $expiration, $scheme, $token));
        }

        $auth_cookie = json_encode($auth_cookie);

        if (false === $auth_cookie) {
            return;
        }

        /**
         * @param string $auth_cookie JSON-encoded array of authentication cookie values.
         * @param int $expiration The time when the authentication cookie expires as a UNIX timestamp.
         * @param int $old_user_id User ID.
         * @param string $scheme Authentication scheme. Values include 'auth' or 'secure_auth'.
         * @param string $token User's session token to use for the latest cookie.
         */
        do_action('set_user_switching_cookie', $auth_cookie, $expiration, $old_user_id, $scheme, $token);

        $scheme = 'logged_in';

        /**
         * @param string $olduser_cookie The old user cookie value.
         * @param int $expiration The time when the logged-in authentication cookie expires as a UNIX timestamp.
         * @param int $old_user_id User ID.
         * @param string $scheme Authentication scheme. Values include 'auth' or 'secure_auth'.
         * @param string $token User's session token to use for this cookie.
         */
        do_action('set_olduser_cookie', $olduser_cookie, $expiration, $old_user_id, $scheme, $token);

        /**
         * @param bool $send Whether to send auth cookies to the client.
         */
        if (!apply_filters('user_switching_send_auth_cookies', true)) {
            return;
        }

        setcookie($auth_cookie_name, $auth_cookie, $expiration, SITECOOKIEPATH, COOKIE_DOMAIN, $secure_auth_cookie, true);
        setcookie(BUILDERIUS_USER_SWITCHING_OLDUSER_COOKIE, $olduser_cookie, $expiration, COOKIEPATH, COOKIE_DOMAIN, $secure_olduser_cookie, true);
    }

    /**
     * @param bool $clear_all Optional. Whether to clear the cookies (as opposed to just popping the last user off the end). Default true.
     * @return void
     */
    public static function clearOldUserCookie($clear_all = true)
    {
        $auth_cookie = self::getAuthCookie();
        if (!empty($auth_cookie)) {
            array_pop($auth_cookie);
        }
        if ($clear_all || empty($auth_cookie)) {
            do_action('clear_olduser_cookie');

            /** This filter is documented in user-switching.php */
            if (!apply_filters('user_switching_send_auth_cookies', true)) {
                return;
            }

            $expire = time() - 31536000;
            setcookie(BUILDERIUS_USER_SWITCHING_COOKIE, ' ', $expire, SITECOOKIEPATH, COOKIE_DOMAIN);
            setcookie(BUILDERIUS_USER_SWITCHING_SECURE_COOKIE, ' ', $expire, SITECOOKIEPATH, COOKIE_DOMAIN);
            setcookie(BUILDERIUS_USER_SWITCHING_OLDUSER_COOKIE, ' ', $expire, COOKIEPATH, COOKIE_DOMAIN);
        } else {
            if (self::secureAuthCookie()) {
                $scheme = 'secure_auth';
            } else {
                $scheme = 'auth';
            }

            $old_cookie = end($auth_cookie);

            $old_user_id = wp_validate_auth_cookie($old_cookie, $scheme);
            if ($old_user_id) {
                $parts = wp_parse_auth_cookie($old_cookie, $scheme);

                if (false !== $parts) {
                    self::setOldUserCookie($old_user_id, true, $parts['token']);
                }
            }
        }
    }

    /**
     * @return false|\WP_User False if there's no old user cookie or it's invalid, WP_User object if it's present and valid.
     */
    public static function getOldUser()
    {
        $cookie = self::getOldUserCookie();
        if (!empty($cookie)) {
            $old_user_id = wp_validate_auth_cookie($cookie, 'logged_in');

            if ($old_user_id) {
                return get_userdata($old_user_id);
            }
        }
        return false;
    }

    /**
     * Gets the value of the cookie containing the originating user.
     *
     * @return string|false The old user cookie, or boolean false if there isn't one.
     */
    public static function getOldUserCookie()
    {
        if (isset($_COOKIE[BUILDERIUS_USER_SWITCHING_OLDUSER_COOKIE])) {
            return wp_unslash($_COOKIE[BUILDERIUS_USER_SWITCHING_OLDUSER_COOKIE]);
        } else {
            return false;
        }
    }

    /**
     * Gets the value of the auth cookie containing the list of originating users.
     *
     * @return string[] Array of originating user authentication cookie values. Empty array if there are none.
     */
    public static function getAuthCookie()
    {
        if (self::secureAuthCookie()) {
            $auth_cookie_name = BUILDERIUS_USER_SWITCHING_SECURE_COOKIE;
        } else {
            $auth_cookie_name = BUILDERIUS_USER_SWITCHING_COOKIE;
        }

        if (isset($_COOKIE[$auth_cookie_name]) && is_string($_COOKIE[$auth_cookie_name])) {
            $cookie = json_decode(wp_unslash($_COOKIE[$auth_cookie_name]));
        }
        if (!isset($cookie) || !is_array($cookie)) {
            $cookie = array();
        }
        return $cookie;
    }

    /**
     * @param \WP_User $new_user Optional. The new user's WP_User object.
     * @param \WP_User $old_user Optional. The old user's WP_User object.
     * @return string The URL to redirect to.
     */
    public static function getRedirect(\WP_User $new_user = null, \WP_User $old_user = null)
    {
        if (!empty($_REQUEST['redirect_to'])) {
            $redirect_to = self::removeQueryArgs(wp_unslash($_REQUEST['redirect_to']));
            $requested_redirect_to = wp_unslash($_REQUEST['redirect_to']);
        } else {
            $redirect_to = '';
            $requested_redirect_to = '';
        }

        if (!$new_user) {
            /** This filter is documented in wp-login.php */
            $redirect_to = apply_filters('logout_redirect', $redirect_to, $requested_redirect_to, $old_user);
        } else {
            /** This filter is documented in wp-login.php */
            $redirect_to = apply_filters('login_redirect', $redirect_to, $requested_redirect_to, $new_user);
        }

        return $redirect_to;
    }

    /**
     * @param \WP_User $user The user to be switched to.
     * @return string The required URL.
     */
    public static function switchToUrl(\WP_User $user)
    {
        return wp_nonce_url(
            add_query_arg(
                [
                    'action' => 'builderius_switch_to_user',
                    'user_id' => $user->ID,
                    'nr' => 1,
                ],
                wp_login_url()
            ),
            "switch_to_user_{$user->ID}"
        );
    }

    /**
     * @param \WP_User $user The old user.
     * @return string The required URL.
     */
    public static function switchBackUrl(\WP_User $user)
    {
        return wp_nonce_url(
            add_query_arg(
                [
                    'action' => 'builderius_switch_to_own_user',
                    'nr' => 1,
                ],
                wp_login_url()
            ),
            "switch_to_own_user_{$user->ID}"
        );
    }

    /**
     * @param \WP_User $user The user to be switched off.
     * @return string        The required URL.
     */
    public static function switchOffUrl(\WP_User $user)
    {
        return wp_nonce_url(
            add_query_arg(
                [
                    'action' => 'builderius_switch_to_unauthenticated',
                    'nr' => 1,
                ],
                wp_login_url()
            ),
            "switch_to_unauthenticated_{$user->ID}"
        );
    }

    /**
     * @return string The current URL.
     */
    public static function currentUrl()
    {
        return $_SERVER['HTTP_REFERER'];
    }

    /**
     * @param string $url A URL.
     * @return string The URL with query args removed.
     */
    public static function removeQueryArgs($url)
    {
        if (function_exists('wp_removable_query_args')) {
            $url = remove_query_arg(wp_removable_query_args(), $url);
        }

        return $url;
    }

    /**
     * Returns whether User Switching's equivalent of the 'logged_in' cookie should be secure.
     *
     * This is used to set the 'secure' flag on the old user cookie, for enhanced security.
     *
     * @link https://core.trac.wordpress.org/ticket/15330
     *
     * @return bool Should the old user cookie be secure?
     */
    public static function secureOldUserCookie()
    {
        return (is_ssl() && ('https' === parse_url(home_url(), PHP_URL_SCHEME)));
    }

    /**
     * Returns whether User Switching's equivalent of the 'auth' cookie should be secure.
     *
     * This is used to determine whether to set a secure auth cookie.
     *
     * @return bool Whether the auth cookie should be secure.
     */
    public static function secureAuthCookie()
    {
        return (is_ssl() && ('https' === parse_url(wp_login_url(), PHP_URL_SCHEME)));
    }

    /**
     * Returns whether the current logged in user is being remembered in the form of a persistent browser cookie
     * (ie. they checked the 'Remember Me' check box when they logged in). This is used to persist the 'remember me'
     * value when the user switches to another user.
     *
     * @return bool Whether the current user is being 'remembered'.
     */
    public static function remember()
    {
        /** This filter is documented in wp-includes/pluggable.php */
        $cookie_life = apply_filters('auth_cookie_expiration', 172800, get_current_user_id(), false);
        $current = wp_parse_auth_cookie('', 'logged_in');

        if (!$current) {
            return false;
        }

        // Here we calculate the expiration length of the current auth cookie and compare it to the default expiration.
        // If it's greater than this, then we know the user checked 'Remember Me' when they logged in.
        return (intval($current['expiration']) - time() > $cookie_life);
    }
}